{ TEST program for the ZIP File System unit (ZFS) - Ivn Montes Velencoso

   WWW : http://webs.demasiado.com/freakpascal/
   e-mail : senbei@teleline.es

    Needs an external file called TEST.ZIP which contains a raw 8bit bitmap
    of 320x200 pixels called ZFSLOGO.RAW, and an indexed palette file of
    256 colours called ZFS.PAL. Both are stored usin DEFLATED method.
}

{$IFNDEF GO32V2}
{$STOP Sorry but this example only works for the go32v2 target}
{$ENDIF}

{$ASMMODE intel}   {Intel assembler syntax}

uses crt, dos, zfs, go32;

var
  pack1     :pZIP;                  {ZIP archives}
  ff        :file;                  {file handler}
  pp        :pointer;               {temporal buffer for picture loading}
  pal       :array[0..767] of byte; {to store the image palette}
  i         :longint;               {counter}

  timer     :DWord absolute $40:$6c;  {system clock}
  time1,
  time2     :DWord;                 {to calculate the time spended}


begin
  ClrSCR;

  WriteLN(' ZiP File System - Test Program 01');
  WriteLN('.-------------------------------.');
  WriteLN;

  pack1:=OpenZIP ('TEST.ZIP', NORMALMODE); {Opens a ZIP file in normal mode}
  if IOResult<>0 then  {There is an error!}
   begin
     WriteLN('Error oppening the ZIP file');
     Halt;
   end;
  WriteLN('# ', pack1^.Name, ' opened.');
  WriteLN;

  LoadOMZ(pack1, 'ZFSLOGO.OMZ');

  Assign(ff, 'ZFSLOGO.raw');  {We open the image file}
  reset(ff,1);
  If IOResult<>0 then {There is an error!}
   begin
     WriteLN('Error while openning a file');
     HALT;
   end;

  {If it's on disk we display a message telling so}
  if filerec(ff).mode<>fmZipped then
   begin
     WriteLN('File found outside a ZIP file, so we use it');
     WriteLN('-Press a key-');
     readkey;
   end;

  Getmem(pp, 64000);      {allocates the needed memory for the buffer}

//  seek(ff, 320*110);    {Uncomment to test the seeking optimization}

  time1:=timer;                       {starts the timming}
  BlockRead(ff, pp^, filesize(ff)-filepos(ff));   {Decompress the file into the buffer}
  time2:=timer;                       {stops the timming}
  i:=IOResult;
  if i<>0 then
   begin
     WriteLN('ERROR reading the file : ', i);
     halt;
   end;

  WriteLN;
  WriteLN(' File info');
  WriteLN('.-------.');
  WriteLN('Name             : ', strpas(filerec(ff).name));
  WriteLN('Size             : ', filesize(ff), ' bytes');
  WriteLN('Current Position : ', filepos(ff));
  WriteLN;
  writeLN('-Press a key-');
  readkey;

  Close(ff);                      {Closes the open file}

  Assign(ff,'ZFS.PAL');           {Loads the image palette}
  {$I-}
  Reset(ff,1);
  BlockRead(ff, pal, 768);
  {$I+}
  i:=IOresult;
  if i<>0 then
   begin
     WriteLN('ERROR LOADING THE PALETTE : ',i);
     HALT;
   end;
  Close(ff);

  asm mov ax, 13h; int 10h; end;  {Sets the VGA 320x200x256 standard mode}

  for i:=0 to 255 do              {Sets the image palette}
   begin                          {the palette is stored in reverse order}
     OutPortb($3c8, i);
     OutPortb($3c9, pal[(255-i)*3+0]);
     OutPortb($3c9, pal[(255-i)*3+1]);
     OutPortb($3c9, pal[(255-i)*3+2]);
   end;

  asm                {Flips the buffer into the VGA Ram}
    cld                            {Clears direction flag}
    push es                        {Saves ES selector}
    mov  es,  DosMemSelector       {ES points to the DOS <1MB memory}
    mov  edi, $a0000               {VGA Ram addres}
    mov  esi, DWord PTR PP
    mov  ecx, 64000/4
    rep  movsd                     {copy the buffer}
    pop  es                        {Restore the ES selector}
  end ['ECX','EDI','ESI'];

  readkey;                         {waits for a keypress}
  asm mov ax, 03h; int 10h; end;   {puts the Text mode}


  FreeMem(pp, 64000);      {Frees the buffer memory}

{  CloseZIP(pack1); }       {It will be closed automatically when exit}
  WriteLN('TIME OF UNCOMPRESSION : ', ((time2-time1)/18.2):8:3, ' seconds');
end.
