{ TEST program for the ZIP File System unit (ZFS) - Ivn Montes Velencoso

   WWW : http://webs.demasiado.com/freakpascal/
   e-mail : senbei@teleline.es

    Needs an external file called TEST.ZIP which contains a raw 8bit bitmap
    of 320x200 pixels called ZFSLOGO.RAW, and an indexed palette file of
    256 colours called ZFS.PAL. Both are stored usin DEFLATED method.

   Thanks to Toxic Avenger/Apocalypse for the window setup routines.
}

{$IFNDEF WIN32}
{$STOP Sorry but this example only works for the Win32 target}
{$ENDIF}

{$ASMMODE intel}   {Intel assembler syntax}

{$APPTYPE GUI}

Program TESTW32;

uses Crt, Dos, Windows, ZFS;

var
  pack1     :pZIP;                  {ZIP archives}
  ff        :file;                  {file handler}
  pp        :pointer;               {temporal buffer for picture loading}
  pal       :array[0..767] of byte; {picture's palette}
  i         :longint;               {counter}
  s         :string;

  time1,
  time2     :DWord;

  hWindow   :HWnd;                  {Window handle}
  hHDC      :HDC;                   {Draw Context}
  pbitmapinfo :pointer;
  AMessage  :Msg;
  vidvscr   :pointer;               {Virtual screen}

Const
  Xres = 320;
  Yres = 200;


PROCEDURE InitGraphic;
Var bisize : longint;
    cmask : array[0..2] of longint;
BEGIN
  bisize:=sizeof(BITMAPINFOHEADER);
  getmem(pbitmapinfo, bisize+12);
  fillchar(pbitmapinfo^, bisize+12, 0);

  with BITMAPINFO(pbitmapinfo^) do
   begin {BitmapInfoHeader 16Bit}
     bmiHeader.biSize        :=bisize;
     bmiHeader.biWidth       :=xres;
     bmiHeader.biHeight      :=-yres;
     bmiHeader.biPlanes      :=1;
     bmiHeader.biBitCount    :=16;
     bmiHeader.biCompression :=BI_BITFIELDS;
   end;

  cmask[0]:=$F800;                       //Bit-Positions R G B 16Bit
  cmask[1]:=$07E0;
  cmask[2]:=$001F;
  move(cmask,ptr(0,longint(pbitmapinfo)+bisize)^,sizeof(cmask));
END;


PROCEDURE BlitGraphic(source : pointer);
BEGIN
  if (hHDC=0) then exit;

  SetDIBitsToDevice(hHDC, 0, 0, xres, yres, 0, 0, 0,
                    yres, source, bitmapinfo(pbitmapinfo^),
                    DIB_RGB_COLORS);
END;


FUNCTION WindowProc(Window: HWnd; AMessage, WParam, LParam: Longint): Longint; stdcall; export;
BEGIN
  WindowProc:=0;

  case AMessage of
    wm_paint: begin
                BlitGraphic(vidvscr);
                Exit;
    end;
    wm_Destroy: begin   { Closed ??!}
                  ReleaseDC(hWindow,hHDC);
                  hHDC:=0;
                  PostQuitMessage(0);
                  Exit;
    end;
  end;

  WindowProc:=DefWindowProc(Window, AMessage, WParam, LParam);
END;


FUNCTION WinRegister: Boolean; {Register Window Class}
Var WindowClass: WndClass;
BEGIN
  WindowClass.Style:=cs_ByteAlignClient;
  WindowClass.lpfnWndProc:=WndProc(@WindowProc);
  WindowClass.cbClsExtra:=0;
  WindowClass.cbWndExtra:=0;
  WindowClass.hInstance:=system.MainInstance;
  WindowClass.hIcon:=LoadIcon(0, idi_Application);
  WindowClass.hCursor:=LoadCursor(0, idc_Arrow);
  WindowClass.hbrBackground:=GetStockObject(BLACK_BRUSH);
  WindowClass.lpszMenuName:=nil;
  WindowClass.lpszClassName:='ZFSTEST'; {Class Name}

  WinRegister:=RegisterClass(WindowClass)<>0;
END;


FUNCTION WinCreate: HWnd;  {Create Window Class}
Var
  hWindow: HWnd;
  r: rect;
  style, styleex : longint;
BEGIN
  r.top   :=CW_USEDEFAULT;
  r.left  :=CW_USEDEFAULT;
  r.right :=xres;
  r.bottom:=yres+GetSystemMetrics(SM_CYCAPTION)-GetSystemMetrics(SM_CYBORDER);

  style:=WS_OVERLAPPED or WS_SYSMENU;
  styleex:=0;

  hWindow:=CreateWindowEx(styleex, 'ZFSTEST', 'ZFS TEST for Win32', style, r.left, r.top, r.right, r.bottom, 0, 0, system.MainInstance, nil);
                                   {Class}    {Window Name}
  if hWindow<>0 then
   begin
     ShowWindow(hWindow, CmdShow);
     UpdateWindow(hWindow);
     hHDC:=GetDC(hWindow);
   end;

  WinCreate:=hWindow;
END;



begin
  pack1:=OpenZIP ('TEST.ZIP', NORMALMODE); {Opens a ZIP file in normal mode}
  if IOResult<>0 then  {There is an error!}
   begin
     MessageBox(0, 'Error oppening ZIP file', nil, mb_Ok);
     Halt;
   end;

  Assign(ff, 'ZFSLogo.raw');  {We open a file}
  reset(ff,1);
  If IOResult<>0 then {There is an error!}
   begin
     MessageBox(0, 'Error while oppening the picture', nil, mb_Ok);
     Halt;
   end;

  LoadOMZ(pack1, 'ZFSLOGO.OMZ');

  {If it's on disk we display a message showing so}
  if filerec(ff).mode<>fmZipped then
   begin
     MessageBox(0, 'File found outside the zip file, so we use it', nil, mb_Ok);
   end;

  Getmem(pp, 64000);      {allocates the needed memory for the buffer}

//  seek(ff, 320*110);    {Uncomment to test the seeking optimization}

  time1:=gettickcount;                       {starts the timming}
  BlockRead(ff, pp^, 64000);   {Decompress the file into the buffer}
  time2:=gettickcount;                       {stops the timming}
  if IOResult<>0 then
   begin
     MessageBox(0, 'Error reading the picture file', nil, mb_Ok);
     Halt;
   end;

  Close(ff);

  Assign(ff,'ZFS.PAL');           {Loads the image palette}
  {$I-}
  Reset(ff,1);
  BlockRead(ff, pal, 768);
  {$I+}
  i:=IOresult;
  if i<>0 then
   begin
     MessageBox(0, 'Error loading palette', nil, mb_Ok);
     Halt;
   end;
  Close(ff);

  if not WinRegister then
   begin
     MessageBox(0, 'Register Window Class failed', nil, mb_Ok);
     Halt(6);
   end;

  hWindow:=WinCreate;
  if (longint(hWindow)=0) then
   begin
     MessageBox(0, 'Window Create failed', nil, mb_Ok);
     Halt(6);
   end;

  InitGraphic;
  getmem(vidvscr,xres*yres*2);

  asm
     mov  ecx, xres*yres {Calculate the lenght of the bitmap}
     mov  esi, DWord PTR PP        {ESI=Original picture (8bit)}
     mov  edi, DWord PTR VidVScr   {EDI=Destination pointer (16bit)}
     lea  edx, PAL

    @@loopA:
      xor   ebx, ebx
      mov   bl, [esi]     { al:=byte(pp^)}
      inc   esi           { inc(pp, 1)}

      neg   ebx
      add   ebx, 255      {Palette is stored in reverse order}
      mov   eax, ebx
      shl   ebx, 1
      add   ebx, eax      {EBX:=(255-CurrentColor)*3}

      mov   al, byte ptr [edx+ebx+2] {b}
      shl   al, 2
      mov   ah, byte ptr [edx+ebx+1] {g}
      mov   bl, byte ptr [edx+ebx]   {r}
      shl   bl, 2
      shr   ax, 3
      and   bl,11111000b
      or    ah,bl        { 8bit Indexed palette -> 16bit pixel done}

      mov   [edi], ax    { word(vidvscr^):=ax}
      add   edi, 2       { inc(vidvscr,2)}

      dec   ecx
      jnz   @loopA
  end ['EAX','EBX','ECX','EDX','EDI','ESI'];


  while GetMessage(@AMessage,0,0,0) do  {Shows the image until a key is pressed}
   begin
     sleep(50);
     TranslateMessage(AMessage);
     DispatchMessage(AMessage);
   end;


  freemem(vidvscr,xres*yres*2);

  FreeMem(pp, 64000);      {Frees the buffer memory}

{  CloseZIP(pack1); }       {It will be closed automatically when exit}

  str( ((time2-time1)/1000.0):0:3, s);
  s:='TIME OF DECOMPRESSION : '+s+' seconds'+#0;
  MessageBox(0, @s[1], 'INFORMATION', mb_ok);
end.
